"""
Name: timer.py
Author: Xuewen Zhang
Date:at 28/04/2024
version: 1.0.0
Description: my own toolbox for research projects
"""
import time
from rich import print as rprint

def timer(f):
    def wrapper(*args, **kwargs):
        start = time.time()
        ret = f(*args, **kwargs)
        end = time.time()
        rprint(f":stopwatch:  Function: [cyan]{f.__name__}[/cyan] ==> Time elapsed: [orange]{end-start:.3f}[/orange] s")
        return ret
    return wrapper


"""
Name: progressbar.py
Author: Xuewen Zhang
Date:at 28/04/2024
version: 1.0.0
Description: my own toolbox for research projects
"""

from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    TaskID,
    TextColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
    TimeElapsedColumn,
    TaskProgressColumn,
    MofNCompleteColumn,
    SpinnerColumn,
    RenderableColumn,
)


class progressbar:

    def probar1():
        """ 
        Progress bar style 1:
            description: 5% =====----------------- 5/100 unit time elapsed < time remaining
        """
        progress = Progress(
            TextColumn(":hourglass_done: [progress.description]{task.description}"),
            TaskProgressColumn(),                              # percentage
            BarColumn(),
            MofNCompleteColumn(),                              # i/total
            TextColumn("[progress.unit]{task.fields[unit]}"),  # add unit
            TimeElapsedColumn(),                               # time elapsed
            TextColumn("<"),                                   # add a '<' symbol
            TimeRemainingColumn(),                             # time remaining
        )
        return progress


    def probar2():
        """ 
        Progress bar style 2:
            description: 5% =====----------------- 5/100 unit time elapsed < time remaining loss：0.00000
        """
        progress = Progress(
            TextColumn(":hourglass_done: [progress.description]{task.description}"),
            # SpinnerColumn(),
            TaskProgressColumn(),                              # percentage
            BarColumn(),                                       # progress bar
            MofNCompleteColumn(),                              # i/total
            TextColumn("[progress.unit]{task.fields[unit]}"),  # add unit
            TimeElapsedColumn(),                               # time elapsed
            TextColumn("<"),                                   # add a '<' symbol
            TimeRemainingColumn(),                             # time remaining
            TextColumn("loss:[progress.loss]{task.fields[loss]:.5f}"), # add wanted information
        )
        return progress
    
    def probar3(*args):
        """ 
        Progress bar style 3:
            description: 5% =====----------------- 5/100 unit time elapsed < time remaining args1：0.00000 args2: 0.00000 ....
        """
        text = []
        for arg in args:
            text.append(f"{arg}:[progress.{arg}]{{task.fields[{arg}]:.5f}}")
            
        progress = Progress(
            TextColumn(":hourglass_done: [progress.description]{task.description}"),
            # SpinnerColumn(),
            TaskProgressColumn(),                              # percentage
            BarColumn(),                                       # progress bar
            MofNCompleteColumn(),                              # i/total
            TextColumn("[progress.unit]{task.fields[unit]}"),  # add unit
            TimeElapsedColumn(),                               # time elapsed
            TextColumn("<"),                                   # add a '<' symbol
            TimeRemainingColumn(),                             # time remaining
            *(TextColumn(text_content) for text_content in text),  
        )
        return progress


if __name__ == "__main__":

    with progress:
        task1 = progress.add_task(description="[red]Downloading...", total=100, loss=0.0)
        i = 0
        while not progress.finished:
            i += 1
            loss = i * 0.001
            progress.update(task1, loss=loss, advance=1)
            time.sleep(0.02)


"""
Name: dirtool.py
Author: Xuewen Zhang
Date:at 03/05/2024
version: 1.0.0
Description: my own toolbox for research projects
"""


import os
import re
import shutil
from rich import print as rprint


class dirtool(object):
    def __init__(self):
        self.author = 'Xuewen Zhang'
        
    
    def create_new_folder_with_max_number(self, directory, prefix=''):
        """
            In the given directory, create a new folder that have a larger number than folders already exist.
            For example: In the directory, have folders mx1, mx2, then create folder mx3
            directory: given path
            prefix: the beginning part of the filenames that you want to check for, e.g. 'exp1', 'exp2' then prefix='exp'
        """
        # check if exit the path, if not, create it.
        self.makedir(directory)

        # Get a list of all folders in the directory
        folders = [folder for folder in os.listdir(directory) if os.path.isdir(os.path.join(directory, folder))]

        # Filter out folders that match the specified pattern
        pattern = re.compile(rf"{prefix}(\d+)")
        matching_folders = [folder for folder in folders if pattern.match(folder)]

        # Extract numbers from matching folder names
        numbers = [int(pattern.match(folder).group(1)) for folder in matching_folders]

        # Find the maximum number
        max_number = max(numbers) if numbers else 0

        # Create a new folder with the maximum number incremented by 1
        new_folder_name = f"{prefix}{max_number + 1}"
        new_folder_path = os.path.join(directory, new_folder_name)
        os.makedirs(new_folder_path)

        return new_folder_name, new_folder_path
    

    def makedir(self, *args):
        """Create the directory"""
        for item in args:
            if not os.path.exists(item):
                os.makedirs(item)
        return
                
                
    def removedir(self, *args):
        """Delete the directory"""
        for item in args:
            shutil.rmtree(item)
        return 
                

    def current_dir(self):
        """Get the path of current file"""
        current_directory = os.path.dirname(os.path.realpath(__file__))
        return current_directory
    

    def copy_file(self, destination_directory, *args):
        """
            Copy the source_file to the destination directory, and generate an info.txt give the original copied file path
            destination_directory (str): destination directory
            args (str): source_file
        """
        for source_file in args:
            try:
                self.makedir(destination_directory)

                # Copy the file to the destination directory
                copied_file_name = os.path.basename(source_file)
                destination_file = os.path.join(destination_directory, copied_file_name)
                shutil.copy(source_file, destination_file)

                # Generate a text file with the content of the original file path and copied file name
                original_file_name = os.path.splitext(copied_file_name)[0]  # Extract filename without extension
                txt_file_path = os.path.join(destination_directory, f"{original_file_name}_info.txt")
                with open(txt_file_path, "w") as txt_file:
                    txt_file.write(f"Original file path: {source_file}\n")
                    txt_file.write(f"Copied file name: {copied_file_name}")

                rprint(f"[cyan] :clipboard: File [white]'{source_file}'[/white] copied to [white]'{destination_directory}'[/white] as [white]'{copied_file_name}'[/white] with [white]'{original_file_name}_info.txt'[/white] [green]successfully[/green].")
            except FileNotFoundError:
                raise ValueError(f"Source file {source_file} not found.")
            except PermissionError:
                raise ValueError("Permission denied.")
            except Exception as e:
                raise ValueError(f"An error occurred: {e}")


"""
Name: datatool.py
Author: Xuewen Zhang
Date:at 28/04/2024
version: 1.0.0
Description: my own toolbox for research projects
"""

import os
import time
import json
import torch
import numpy as np
import casadi as cs
from rich import print as rprint


class datatool(object):
    def __init__(self):
        self.author = 'Xuewen Zhang'
        self.scale_tracker_reset()
        
        
    ## ---------------------- Data processing ---------------------- ##    
    # Change data type or device
    def toArray(self, *args):
        """Turn data to array"""
        data_list = []
        for data in args:
            if isinstance(data, np.ndarray):
                data_list.append(data)
            else:
                data_list.append(np.array(data))
        if len(data_list) == 1:
            data_list = data_list[0]
        return data_list


    def toList(self, *args):
        """Convert NumPy arrays/Torch tensors to Python lists"""
        data = []
        for x in args:
            if isinstance(x, list):
                data.append(x)
            else:
                data.append(x.tolist())
        if len(data) == 1:
            data = data[0]
        return data


    def toTorch(self, *args):
        """Turn data to array"""
        data_list = []
        for data in args:
            if isinstance(data, torch.Tensor):
                data_list.append(data)
            else:
                data_list.append(torch.FloatTensor(data))
        if len(data_list) == 1:
            data_list = data_list[0]
        return data_list


    def toDevice(self, device=None, *args):
        data_all = []
        if device is None:
            device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        for data in args:
            if torch.is_tensor(data):
                data = data.to(device)
            else:
                raise ValueError("Data/model is not torch type.")
            data_all.append(data)
        if len(data_all) == 1:
            data_all = data_all[0]
        return data_all
    
    
    # data scale
    def _scale_tracker(self, idx, value):
        """ Track if the idx and value shows up before """
        value_tuple = tuple(value)
        
        if (tuple(idx), value_tuple) in self.scale_track_dict:
            # If the data has been seen before, return 1
            return True
        else:
            # If the data has not been seen before, add it to the dictionary and return 0
            self.scale_track_dict[(tuple(idx), value_tuple)] = True
            return False
        
        
    def scale_tracker_reset(self):
        """ Reset the scale tracker """
        self.scale_track_dict = {}
        return
            
        
    def scale(self, sc1, sc2, method='minmax', datatype='array', *args):
        """
            Scale the data using 'minmax' or 'norm' methods
        Args:
            sc1 (array or tensor)(1, dim) or (dim,): if minmax, sc1 is the min value of x; if norm, sc1 is the mean value of x
            sc2 (array or tensor)(1, dim) or (dim,): if minmax, sc2 is the max value of x; if norm, sc2 is the std value of x
            method (str, optional): scale methods. Defaults to 'minmax', including 'minmax' and 'norm'.
            datatype (str, optional): data type. Defaults to 'array', including 'array' and 'tensor'.
            args: contains data x (array or tensor)(1, dim) or (dim,): original data
        Returns:
            x_ (array or tensor)(1, dim) or (dim,): scaled data
        """
        datatype_dict = {'array', 'tensor'}
        method_dict = {'minmax', 'norm'}
        self.checkarg([datatype, datatype_dict], [method, method_dict])

        data_scale  = []
        i = 0
        for x in args:
            i += 1
            if method == 'minmax':
                x_ = (x - sc1)/(sc2 - sc1)
                # check if have x_max - x_min == 0, if have, replace to zeros
                zero_idx, zero_dim = self.find0idx(datatype, sc2 - sc1)
                if len(zero_idx) != 0:
                    if zero_dim == 0:
                        x_[zero_idx] = 0 
                    else:
                        x_[:, zero_idx] = 0
                    constant_value = x[zero_idx] if zero_dim == 0 else x[0, zero_idx]
                    
                    if not self._scale_tracker(zero_idx, constant_value):
                        # only print once for the same constant value
                        rprint(f"[red] :warning: [cyan]'{i}'[/cyan] input data have constant value, index: [cyan]'{zero_idx}'[/cyan]: value: [cyan]{constant_value}[/cyan], replaced to [cyan]'0'[/cyan].")
            elif method == 'norm':
                x_ = (x - sc1)/sc2
            data_scale.append(x_)
            
        if len(data_scale) == 1:
            data_scale = data_scale[0]
        return data_scale
    
    
    def unscale(self, sc1, sc2, method='minmax', datatype='array', *args):
        """
            Unscale the data using 'minmax' or 'norm' methods
        Args:
            sc1 (array or tensor)(1, dim) or (dim,): if minmax, sc1 is the min value of x; if norm, sc1 is the mean value of x
            sc2 (array or tensor)(1, dim) or (dim,): if minmax, sc2 is the max value of x; if norm, sc2 is the std value of x
            method (str, optional): scale methods. Defaults to 'minmax', including 'minmax' and 'norm'.
            datatype (str, optional): data type. Defaults to 'array', including 'array' and 'tensor'.
            args: contains data x_ (array or tensor)(1, dim) or (dim,): scaled data
        Returns:
            x (array or tensor)(1, dim) or (dim,): unscaled data
        """
        datatype_dict = {'array', 'tensor'}
        method_dict = {'minmax', 'norm'}
        self.checkarg([datatype, datatype_dict], [method, method_dict])

        data = []
        i = 0
        for x_ in args:
            i += 1
            if method == 'minmax':
                x = x_ * (sc2 - sc1) + sc1
                # check if have x_max - x_min == 0, if have, replace to constant value
                zero_idx, zero_dim = self.find0idx(datatype, sc2 - sc1)
                if len(zero_idx) != 0:
                    constant_value = sc1[zero_idx] if zero_dim == 0 else sc1[:, zero_idx]
                    constant_value_slice = constant_value.copy() if zero_dim == 0 else constant_value[0].copy()
                    if zero_dim == 0:
                        x[zero_idx] = constant_value  
                    else:
                        x[:, zero_idx] = constant_value
                    if not self._scale_tracker(zero_idx, constant_value_slice):
                        # only print once for the same constant value
                        rprint(f"[red] :warning: [cyan]'{i}'[/cyan] input data have constant value, index: [cyan]'{zero_idx}'[/cyan]: unscale to constant value: [cyan]{constant_value_slice}[/cyan].")
            elif method == 'norm':
                x = x_ * sc2 + sc1
            data.append(x)
            
        if len(data) == 1:
            data = data[0]
        return data
    

    # other utils
    def mse(self, data1, data2):
        """Compute mse of data1 and data2"""
        if data1.shape != data2.shape:
            raise ValueError(f'data1 shape {data1.shape} != data2 shape {data2.shape}')
        else:
            mse = np.linalg.norm(data1 - data2, ord=2)  # 2 norm
        return mse
    
    
    def data_to_step(self, t=None, *args):
        """
            Transform the sequential data to step data
            args: x (data_size, x_dim): used for the sequential data such as control inputs
            t (data_size): specially for time data
        """ 
        data = []
        for x in args:
            x = x.T
            dim, n = x.shape
            x_step = np.zeros((dim, 2*n))

            for i in range(n):
                x_step[:, 2 * i] = x[:, i]
                x_step[:, 2 * i + 1] = x[:, i]
            
            data.append(x_step.T)

        if len(data) == 1:
            data = data[0]
        
        if t is not None:
            t_step = np.zeros(2*len(t))
            for i in range(len(t)):
                t_step[2 * i] = t[i]
                t_step[2 * i + 1] = t[i]
            ts = t[1]-t[0]
            t_step = t_step[1:]
            t_step = np.concatenate((t_step, np.array([t_step[-1] + ts])))
            return t_step, data 
        else:
            return data

    
    def checkarg(self, *args):
        """
            Check if the function inputs are in the feasible sets
        Args:
            args (list|[variable, variable_dict]): values of different variable inputs and feasible sets.
        """
        for [variable, feasible_dict] in args:
            if variable not in feasible_dict:
                raise ValueError(f"Input {variable} is not supported, should be {feasible_dict}.")
    
    
    def find0idx(self, datatype='array', *args):
        """
            Find the zeros index of the data x.
        Args: 
            args: x (array or tensor)(1, dim) or (dim,): data
            datatype (str, optional): data type. Defaults to 'array', including 'array' and 'tensor'.
        Returns:
            zero_idx (list): zeros index
        """
        datatype_dict = {'array', 'tensor'}
        self.checkarg([datatype, datatype_dict])
        
        if datatype == 'array':
            where = np.where
        elif datatype == 'tensor':
            where = torch.where
        
        idx, dim = [], []
        for x in args:
            dim_size = len(x.shape)
            zero_dim = dim_size - 1
            zero_idx = where(x == 0)[zero_dim]
            idx.append(zero_idx)
            dim.append(zero_dim)
        if len(idx) == 1:
            idx = idx[0]
            dim = dim[0]
        return idx, dim

        
    
    ## ---------------------- Load data ---------------------- ##
    def loadtxt(self, loaddir, *args):
        data = []
        for name in args:
            data.append(np.loadtxt(loaddir + '%s.txt' % name))
        if len(data) == 1:
            data = data[0]
        return data
    

    def loadpt(self, loaddir, device=torch.device('cpu'), *args):
        """
            Load pt files using torch
            loaddir: save path
            device: cpu or gpu, {'cuda:0'}
            *args: file name in string
        """
        data = []
        for name in args:
            data.append(torch.load(loaddir + '%s.pt' % name, map_location=device))
        if len(data) == 1:
            data = data[0]
        return data


    def load_config(self, savedir='', *args):
        """args: the str of the load josn file name"""
        data_list = []
        for name in args:
            config = open(savedir + '%s.json' % name)
            data = json.load(config)
            data_list.append(data)
        if len(data_list) == 1:
            data_list = data_list[0]
        return data_list


    ## ---------------------- Save data ---------------------- ##
    def savetxt(self, savedir, **kwargs):
        """Save the data into txt files"""
        for name, item in kwargs.items():
            if isinstance(item, torch.Tensor):
                item = self.toArray(item.cpu())
            if isinstance(item, list):
                item = self.toArray(item)
            if item.ndim == 0:
                item = np.array([item])
            np.savetxt(savedir + '%s.txt' % name, item)


    def save_datatoconfig(self, save_dir='', **kwargs):
        """Write the dict into txt"""
        for name, data in kwargs.items():
            file = open(save_dir + '%s.json' % name, 'w')
            file.write('{\n')

            count = 0

            for key, value in data.items():
                count += 1
                if count != len(data.keys()):
                    file.write(f'"{key}": {value},\n')
                else:
                    file.write(f'"{key}": {value}\n')

            file.write('}\n')
            file.close()


    def save_config(self, args, savedir, name):
        """Save the config into json file"""
        json_dir_val = os.path.join(savedir, "{}.json".format(name))
        with open(json_dir_val, 'w') as fp:
            json.dump(args, fp, indent=4, default=self.toList)


    def add_line_to_file(self, file_path, line):
        """Add one line sentence to certain txt"""
        with open(file_path, 'a') as f:
            f.write(line + '\n')
            
            
    def save_model(self, model, savedir, train_loss, train_loss_all, val_loss=None, val_loss_all=None):
        """
        Used for saving the best training model and validation model
        Args:
            model: the NN model
            train_loss: each train epoch loss
            val_loss:each validation epoch loss
        Returns: None
        """
        # saves model weights if loss is the lowest ever
        train_min = np.min(train_loss_all)
        if train_loss <= train_min:
            torch.save({"nn_model": model.state_dict()}, savedir+'train_model.pt')
        
        if val_loss is not None and val_loss_all is not None:
            val_min = np.min(val_loss_all)
            if val_loss <= val_min:
                torch.save({"nn_model": model.state_dict()}, savedir+'val_model.pt')
                
                




"""
Name: mathtool.py
Author: Xuewen Zhang
Date:at 28/04/2024
version: 1.0.0
Description: my own toolbox for research projects
"""

import numpy as np
import torch

class mathtool(object):
    def __init__(self):
        self.author = 'Xuewen Zhang'
        
        
    def hankel(self, L, datatype='array', *args):
            """
            Construct the Hankel matrix
                datatype (str): 'array' or 'tensor'
                args: x: data sequence (data_size, x_dim)
                L: row dimension of the hankel matrix
                T: data samples of data x
                return H(x): hankel matrix of x  H(x): (x_dim*L, T-L+1)
                H(x) = [x(0)   x(1) ... x(T-L)
                        x(1)   x(2) ... x(T-L+1)
                        .       .   .     .
                        .       .     .   .
                        .       .       . .
                        x(L-1) x(L) ... x(T-1)]
            """
            if datatype not in ['array', 'tensor']:
                raise ValueError(f"{datatype} not exist, datatype should be 'array' or 'tensor'.")

            hankeldata = []
            for x in args:
                T, x_dim = x.shape
                Hx = torch.zeros((L*x_dim, T-L+1), device=x.device) if datatype == 'tensor' else np.zeros((L*x_dim, T-L+1))
                for i in range(L):
                    Hx[i*x_dim:(i+1)*x_dim, :] = x[i:T-L+i+1, :].transpose(0, 1) if datatype == 'tensor' else x[i:T-L+i+1, :].transpose(1, 0)  # x need transpose to fit the hankel dimension
                hankeldata.append(Hx)
                
            if len(hankeldata) == 1:
                hankeldata = hankeldata[0]
            return hankeldata
        
        


"""
Name: log.py
Author: Xuewen Zhang
Date:at 03/05/2024
version: 1.0.0
Description: my own toolbox for research projects
"""

import wandb
import time

def new_wandb_proj(project_name, exp_name=time.strftime('%y%m%d'), config=None, project_notes=''):
    """
    Create a new project on wandb
    Args:
        project_name (str): the name of the project
        config (dict): the configuration of the project inculding the hyperparameters like {'lr': 0.01, 'batch_size': 32}
        project_notes (str): the description of the project
    Returns: None
    """
    wandb.init(project=project_name, name=exp_name, notes=project_notes, config=config)
    ## then use wandb.log() to log the data
    # wandb.log({'loss': 0.5, 'accuracy': 0.9})
    # ytable = wandb.Table(data=y_dpc, columns=['y' + str(i) for i in range(self.y_dim)])
    # wandb.log({'ytable': ytable})
    # yu_img = wandb.Image(fig1, caption='yu_deepc')
    # wandb.log({'yu_deepc': yu_img})
    return


"""
Name: nntool.py
Author: Xuewen Zhang
Date:at 03/05/2024
version: 1.0.0
Description: my own toolbox for research projects
"""

import os
import torch
import numpy as np
import casadi as cs
from rich import print as rprint


class nntool():
    def __init__(self):
        """
            Neural network related tools
        """
        self.name = 'Xuewen'
        self.flags = {'initialize': False}


    def init(self, model):
        """ Initialize the model """
        self.model = model
        self.model_param = dict(self.model.named_parameters())
        self.model_param_name = list(self.model_param.keys())
        self.flags['initialize'] = True
        return
    
    
    def load_model(self, paramdir, device=torch.device('cpu'), model=None, checkpoint_name='nn_model'):
        """
            Load the saved model
            paramdir: the path of the saved model pt files
            device: the device to load the model, cpu or gpu {'cuda:0'}
            kwargs: the model parameters
        """
        checkpoint = torch.load(paramdir, map_location=device)
        if model is None:
            if not self.flags['initialize']:
                raise ValueError('Please initialize the nntool (self.init(model)) first if do not provide the model.')
            self.model.load_state_dict(checkpoint[checkpoint_name])
            model = self.model
        else:
            model.load_state_dict(checkpoint[checkpoint_name])
        rprint(":inbox_tray: Model loaded [green]successfully[/green].")
        return model
    

    def save_model(self, savedir, train_loss, val_loss, train_loss_all, val_loss_all, model=None):
        """
        Used for saving the best training model and validation model
        Args:
            model: the NN model
            train_loss: each train epoch loss
            val_loss:each validation epoch loss
        Returns: None
        """
        if model is None:
            if not self.flags['initialize']:
                raise ValueError('Please initialize the nntool (self.init(model)) first if do not provide the model.')
            model = self.model
        # Validation sets are used to decide the parameters of the model
        train_min = np.min(train_loss_all)
        val_min = np.min(val_loss_all)

        # saves model weights if loss is the lowest ever
        if val_loss <= val_min:
            torch.save({"nn_model": model.state_dict()}, savedir+'val_model.pt')
        if train_loss <= train_min:
            torch.save({"nn_model": model.state_dict()}, savedir+'train_model.pt')
        return


    def save_parameters(self, savedir, model=None):
        """
            Save model parameters
            model: neural network model
            savedir: save path for parameters
        """
        if model is None:
            if not self.flags['initialize']:
                raise ValueError('Please initialize the nntool (self.init(model)) first if do not provide the model.')
            model = self.model
            model_param = self.model_param
        else:
            model_param = dict(model.named_parameters())
        
        if not os.path.exists(savedir):
            os.makedirs(savedir)

        for name, value in model_param.items():
            np.savetxt(savedir + '{}.txt'.format(name), value.detach().numpy())
        rprint(':page_facing_up: Model parameters saved as TXTs [green]successfully[/green].')
        return


    def model_casadi(self, input_dim, activation_fcn=None, model=None):
        """
            Construct the well-trained model using casadi and wrap it as a casadi Function for control optimization
            May need to change the structure according to the specific task
            input_dim: the dimension of the input layer
            activation_fcn: 'relu', 'tanh', if None, then no activation functions
            return: self.model_cas a function   output = self.model_cas(input)
        """
        if model is None:
            if not self.flags['initialize']:
                raise ValueError('Please initialize the nntool (self.init(model)) first if do not provide the model.')
            model_param = self.model_param
            model_param_name = self.model_param_name
        else:
            model_param = dict(model.named_parameters())
            model_param_name = list(model_param.keys())
            
        input_var = cs.SX.sym('input', input_dim)
        layer_num = int(len(self.model_param_name)/2)

        inputs = input_var
        for i in range(layer_num):
            fc_weight = np.array(model_param[model_param_name[i]].detach().numpy())
            fc_bias = np.array(model_param[model_param_name[i+1]].detach().numpy())
            output = cs.mtimes(fc_weight, input) + fc_bias
            if i != layer_num-1:                        # last layer no activation function
                if activation_fcn == 'relu':            # different activation functions
                    output = cs.fmax(output)
                elif activation_fcn == 'tanh':
                    output = cs.tanh(output)
            inputs = output                              # update input for the next layer

        # Create casadi function
        self.model_cas = cs.Function('model', [input_var], [output])
        return self.model_cas
    

    








"""
Name: QYtool.py
Author: Xuewen Zhang
Date:at 28/04/2024
version: 1.0.0
Description: My own toolbox for research projects
"""

from .datatool import datatool
from .dirtool import dirtool
from .nntool import nntool
from .mathtool import mathtool
from .progressbar import progressbar
from .wandblog import new_wandb_proj
from .timer import timer

from rich import print as rprint
from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    TaskID,
    TextColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
    TimeElapsedColumn,
    TaskProgressColumn,
    MofNCompleteColumn,
    SpinnerColumn,
    RenderableColumn,
)

datatool = datatool()
dirtool = dirtool()
mathtool = mathtool()
nntool = nntool()

